# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Collection of functions to assist with data typing."""

from ayx_learn.utils.constants import ColumnTypes

import numpy as np

import pandas as pd


def to_numeric(series):
    """Convert a series to numeric values."""
    series = pd.to_numeric(series, errors="coerce")

    if series.dtype != np.float64:
        series = series.apply(val_to_float)

    return series


def to_categorical(series):
    """Convert a series to categorical values."""
    series = series.astype("category")
    series = series.apply(str)
    return series


def to_boolean(series):
    """Convert a series to boolean values."""
    col_to_bool = np.vectorize(val_to_bool)
    return col_to_bool(series).astype(float)


def to_id(series):
    """Convert a series to an ID type."""
    return series


def val_to_float(x):
    """Convert a single number to a floating point value."""
    try:
        return float(x)
    except Exception:
        return np.nan


def val_to_bool(val):
    """Convert a single number to a boolean value."""
    val = str(val).strip().upper()

    try:
        return {
            "1": True,
            "1.0": True,
            "TRUE": True,
            "T": True,
            "Y": True,
            "YES": True,
            "0": False,
            "0.0": False,
            "FALSE": False,
            "F": False,
            "N": False,
            "NO": False,
        }[val]
    except KeyError:
        return None


typing_conversions = {
    ColumnTypes.NUMERIC: to_numeric,
    ColumnTypes.CATEGORICAL: to_categorical,
    ColumnTypes.BOOLEAN: to_boolean,
    ColumnTypes.ID: to_id,
}



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Impute method transformer."""

from enum import Enum

from ayx_learn.utils.validate import validate_enum

import pandas as pd

import sklearn.base


def mean(df, *args):
    """Calculate fill value in MEAN mode."""
    return df.mean()


def mode(df, *args):
    """Calculate fill value in MODE mode."""
    return df.mode()[0]


def median(df, *args):
    """Calculate fill value in MEDIAN mode."""
    return df.median()


def value(_, val, *args):
    """Calculate fill value in VALUE mode."""
    return val


def custom(df, _, func, *args):
    """Calculate fill value in CUSTOM mode."""
    if not func:
        raise ValueError("Function has not been properly specified.")
    return func(df)


impute_method_dict = {
    "MEAN": mean,
    "MEDIAN": median,
    "MODE": mode,
    "VALUE": value,
    "CUSTOM": custom,
}

# Auto generate the Enum for Impute methods based on the above dict.
# This implicitly forces any additional imputation methods to have
# associated functions
ImputeMethods = Enum("ImputeMethods", " ".join(impute_method_dict.keys()))


class ImputeTransformer(sklearn.base.TransformerMixin):
    """Impute columns in a DataFrame.

    Parameters
    ----------
    colnames: List[str]
        Columns to impute.

    method: ImputeMethods
        Imputation method to apply. MEAN, MODE, MEDIAN, VALUE, or CUSTOM.

    value: float
    """

    def __init__(
        self,
        method: ImputeMethods = ImputeMethods.MEAN,
        value: float = 0.0,
        custom_method=None,
        colname=None,
        inplace=False,
        **kwargs
    ):
        self.method = method
        self.value = value
        self.col_value_dict = {}
        if custom_method:
            self.custom_method = custom_method
        else:
            self.__custom_method = None
        self.colname = colname
        self._inplace = bool(inplace)

    @property
    def method(self):
        """Get imputation method."""
        return self.__method

    @method.setter
    def method(self, method: ImputeMethods):
        """Set imputation method."""
        validate_enum(method, ImputeMethods)
        self.__method = method

    @property
    def value(self):
        """Get value in VALUE imputation method."""
        return self.__value

    @value.setter
    def value(self, value: float):
        """Set value in VALUE imputation method."""
        self.__value = value

    @property
    def custom_method(self):
        """Getter for custom_method used in the CUSTOM imputer mode."""
        return self.__custom_method

    @custom_method.setter
    def custom_method(self, custom_method):
        """Setter for custom_method used in the CUSTOM imputer mode."""
        if not callable(custom_method):
            raise ValueError("Custom method must be callable.")
        self.__custom_method = custom_method

    def transform(self, x: pd.DataFrame, y=None, **kwargs) -> pd.DataFrame:
        """Impute the colnames in the DataFrame.

        Parameters
        ----------
        x : pandas.DataFrame
            Incoming DataFrame.

        Returns
        -------
        pandas.DataFrame
            Imputed DataFrame

        Raises
        ------
        ValueError
            If imputation method is invalid.
        Exception
            If an error occurs, such as invalid colname.
        """
        columns_to_impute = x.columns
        if self.colname is not None:
            columns_to_impute = [self.colname]
        inplace = self._inplace
        if not inplace:
            x = x.copy()
        for column in columns_to_impute:
            try:
                # Extract the column
                data = x[column]

                # Calculate the fill value
                try:
                    value = self.col_value_dict[column]
                except KeyError:
                    raise ValueError("A column is present that was not during fit.")

                # Perform imputation
                if data.dtype.name == "category" and value not in data.cat.categories:
                    data = data.cat.add_categories(value)
                x.loc[:, column] = data.fillna(value)
            except Exception as err:
                raise err
        return x

    def fit(self, x: pd.DataFrame, y=None, **kwargs):
        """Return self."""
        method = self.method
        columns_to_impute = x.columns
        if self.colname is not None:
            columns_to_impute = [self.colname]
        for column in columns_to_impute:
            try:
                # Extract the column
                data = x[column]

                # Calculate the fill value
                value = impute_method_dict[method.name](
                    data, self.value, self.custom_method
                )

                # Save it
                self.col_value_dict[column] = value

            except Exception as err:
                raise err
        return self



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""FeatureModifier used for operating on subsets of DataFrame columns using regex."""

from ayx_learn.pipeline_utils.dataframe_feature_union import DataFrameFeatureUnion
from ayx_learn.transformers.column_selector_transformer import ColumnSelectorTransformer
from ayx_learn.transformers.dataframe_function_transformer import (
    DataFrameFunctionTransformer,
)

import pandas as pd

import sklearn.base
import sklearn.pipeline


# Feature modifier: takes a list of regular expressions to
# indicate columns to apply the transformer to
# If in place is set to true, the original columns are removed,
# if false, the new features are unioned to the
# original data set. The columns not selected by
# col_select are untouched
class FeatureModifier(sklearn.base.TransformerMixin, sklearn.base.BaseEstimator):
    """
    Custom transformer for operating on features using regex.

    Transformer that is designed to operate on Pandas
    DataFrame objects by mapping a sklearn Transformer to a set
    of one or more columns in a Pandas DataFrame.  Several
    DataframeFeatureModifier instances can be chained together in
    an sklearn.Pipeline, returning a pandas Dataframe as output.
    """

    def __init__(
        self,
        transformer,
        col_select=None,
        in_place=True,
        name_scheme="in_place",
        name="",
        drop_unselected=False,
    ):
        """
        Initialize a FeatureModifier.

        Arguments:
            transformer: sklearn.TransformerMixin -- An sklearn transformer
            that will be applied to the columns specified in col_select

        Keyword Arguments:
            col_select : list -- A list containing regular expression
                selectors that will be applied for selecting
                one or more DataFrame columns for the specified
                transformer to operate on (default: {None})
            in_place : bool -- Specifies whether transformed columns
                will replace original columns, or simply be
                appended to original columns with names
                generated according to name_scheme (default: {True})
            name_scheme : str -- "in_place"|"append"|"new"
                "append": The value specified in name will
                be appended to original column name as the name
                of newly transformed columns
                "new" : The value specified in name will
                be used for new columns, with a numeric value,
                separated by underscore, appended to the new column
                name (default: {in_place})
            name : str -- Value used for new columns generated,
                according to values specified for "in_place" and
                "name_scheme" parameters. (default: {""})
        """
        self._drop_unselected = drop_unselected
        self._transformer = transformer

        if not in_place and name_scheme == "in_place":
            raise ValueError(
                f"""Must use different naming scheme than
                {name_scheme} when operating in place."""
            )

        self._name_scheme = name_scheme
        self._name = name

        # If it's in place, then we want to deselect the columns that we are
        # operating on. If it's not in place (i.e. we're appending new features)
        # then we want to keep all the original columns, i.e. deselect None
        if in_place:
            self._unchanged_selector = ColumnSelectorTransformer(
                col_select=col_select, select_not_deselect=False
            )
        else:
            self._unchanged_selector = ColumnSelectorTransformer(
                col_select=None, select_not_deselect=False
            )

        # Select the columns to apply the transformer to
        self._change_selector = ColumnSelectorTransformer(
            col_select=col_select, select_not_deselect=True
        )

        self.steps_list = [
            (
                "2",
                sklearn.pipeline.make_pipeline(
                    self._change_selector,
                    DataFrameFunctionTransformer(self._register_original_col_names),
                    self._transformer,
                    DataFrameFunctionTransformer(self._create_output_df),
                ),
            )
        ]

        if not self._drop_unselected:
            self.steps_list.insert(
                0, ("1", sklearn.pipeline.make_pipeline(self._unchanged_selector))
            )

        # Build the pipeline to run, we will individually run the
        # transformer on the selected columns, and then feature union the result
        # with the unchanged columns
        self._internal_pipeline = DataFrameFeatureUnion(self.steps_list)

    def _create_output_df(self, x, y=None):
        if isinstance(x, pd.DataFrame):
            return x
        return pd.DataFrame(x, columns=self._gen_col_names(x))

    def _register_original_col_names(self, x, y=None):
        self._orig_cols = list(x)
        return x

    def _intersection_empty(self, lst1, lst2):
        return len(self._intersection(lst1, lst2)) == 0

    def _intersection(self, lst1, lst2):
        return [value for value in lst1 if value in lst2]

    def _validate_columns(self, new_cols):
        if not self._intersection_empty(new_cols, self._orig_cols):
            raise ValueError(
                f"""Column names contain duplicates. The following
            column names appear in original and new columns:
            {self._intersection(new_cols, self._orig_cols)}"""
            )

    def _gen_col_names(self, x):
        if self._name_scheme == "in_place":
            return self._orig_cols
        elif self._name_scheme == "append":
            new_cols = [name + self._name for name in self._orig_cols]
            self._validate_columns(new_cols)
            return new_cols
        elif self._name_scheme == "new":
            new_cols = [self._name + "_" + str(ind) for ind in range(len(x[0]))]
            self._validate_columns(new_cols)
            return new_cols
        else:
            raise ValueError("Unknown naming scheme")

    def fit(self, x, y=None, **kwargs):
        """Fit a dataframe."""
        assert isinstance(x, pd.DataFrame)
        return self._internal_pipeline.fit(x, y)

    def transform(self, x, y=None, *args, **kwargs):
        """Transform a dataframe."""
        assert isinstance(x, pd.DataFrame)
        return self._internal_pipeline.transform(x, y)



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Linear regression algorithm."""

from sklearn.linear_model import ElasticNet


class LinearRegression(ElasticNet):
    """Linear regression wrapper class."""

    def __init__(
        self,
        alpha=1.0,
        l1_ratio=0.5,
        fit_intercept=True,
        normalize=True,
        precompute=False,
        max_iter=1000,
        copy_X=True,  # noqa: N803
        tol=0.0001,
        warm_start=False,
        positive=False,
        random_state=10,
        selection="cyclic",
    ):
        """
        Linear regression object constructor.

        The only differences between this implementation and sklearns are:
            1. normalize defaults to True. Normalizing variables is best practice when
            using ElasticNet since it includes weight regularization.

            2. random_state default to 10. Since random_state is not yet exposed
            through the CFML configuration panels, default seeding is necessary to
            guarantee consistency.
        """
        super().__init__(
            alpha,
            l1_ratio,
            fit_intercept,
            normalize,
            precompute,
            max_iter,
            copy_X,
            tol,
            warm_start,
            positive,
            random_state,
            selection,
        )



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Exception definitions for ayx-learn."""


class DataScienceException(Exception):
    """Base exception for data science objects."""

    pass


class OheUnexpectedLevelsException(DataScienceException):
    """Exception for unexpected level in One Hot Encoding."""

    def __init__(self, message, errors):
        self.levels, self.column_name = errors
        super().__init__(message)


class NullValueError(DataScienceException):
    """Exception for a dataframe containing null values."""

    def __init__(self, message, column_name):
        self.column_name = column_name
        super().__init__(message)



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Collection of transformers used in Alteryx."""

from .column_selector_transformer import ColumnSelectorTransformer
from .column_type_transformer import ColumnTypeTransformer
from .drop_na_rows_transformer import DropNaRowsTransformer
from .identity_transformer import IdentityTransformer
from .impute_transformer import ImputeTransformer
from .one_hot_encoder_transformer import OneHotEncoderTransformer

__all__ = [
    "ColumnSelectorTransformer",
    "ColumnTypeTransformer",
    "DropNaRowsTransformer",
    "IdentityTransformer",
    "ImputeTransformer",
    "OneHotEncoderTransformer",
]



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""One hot encoder implementation."""
import logging

from ayx_learn.utils.exceptions import OheUnexpectedLevelsException
from ayx_learn.utils.validate import validate_no_nulls

import numpy as np

import pandas as pd

from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.validation import check_is_fitted

logger = logging.getLogger(__name__)


class OneHotEncoderTransformer(OneHotEncoder):
    """Used to perform one-hot-encoding on a dataframe."""

    def __init__(self, *args, categorical_columns="auto", **kwargs):
        """Construct one hot encoding object."""
        super().__init__(*args, **kwargs, sparse=False)
        self.categorical_columns = categorical_columns
        self._input_columns = []
        self._columns_to_encode = []

    @property
    def categorical_columns(self):
        """Getter for categorical columns."""
        return self.__categorical_columns

    @categorical_columns.setter
    def categorical_columns(self, value):
        if not isinstance(value, list) and value != "auto" and value is not None:
            raise ValueError(
                "categorical_columns must be a list of column names, 'auto', or None."
            )
        self.__categorical_columns = value

    def fit(self, x, *_, **__):
        """Fit one hot encoder."""
        self._input_columns = list(x)

        if self.categorical_columns is None:
            self._columns_to_encode = self._input_columns
        elif self.categorical_columns == "auto":
            self._columns_to_encode = x.select_dtypes(
                include=["category", "object"]
            ).columns.tolist()
        else:
            self._columns_to_encode = self.categorical_columns

        self.categories = [
            sorted(list(set(x[col].values))) for col in self._columns_to_encode
        ]

        try:
            super().fit(x[self._columns_to_encode])
        except ValueError as e:
            logger.exception(e)
            validate_no_nulls(x)
            raise e

        return self

    def transform(self, x, *_, **__):
        """One hot encode new data."""
        check_is_fitted(self, ["categories_"])

        if set(x.columns) != set(self._input_columns):
            err_str = (
                f"Incoming columns don't match fitted columns: {self._input_columns}"
            )
            logger.error(err_str)
            raise ValueError(err_str)
        elif list(x.columns) != list(self._input_columns):
            x = x[self._input_columns]

        try:
            x_encoded = super().transform(x[self._columns_to_encode])
        except ValueError as e:
            logger.exception(e)
            validate_no_nulls(x)
            self._validate_no_unexpected_categories(x)
            raise e

        x_encoded = pd.DataFrame(
            x_encoded, columns=self.get_categorical_feature_names(), index=x.index
        )

        unencoded_columns = [
            col for col in list(x) if col not in self._columns_to_encode
        ]

        if len(unencoded_columns) == 0:
            return x_encoded

        unencoded_df = x[unencoded_columns]

        x_encoded[list(unencoded_df)] = unencoded_df

        return x_encoded

    def get_feature_names_(self):
        """Get the the names of all the output features."""
        check_is_fitted(self, ["categories_"])

        features = self.get_categorical_feature_names()

        unique_names = set(self._input_columns)
        for idx, feature in enumerate(features):
            new_feature = self._get_unique_name(feature, unique_names)
            features[idx] = new_feature
            unique_names.add(new_feature)

        for feature in self._input_columns:
            if feature not in self._columns_to_encode:
                features.append(feature)

        return features

    def get_categorical_feature_names(self):
        """Get the names of the output categorical features."""
        return super().get_feature_names(self._columns_to_encode).tolist()

    def get_link(self):
        """Get the link between original column names and encoded ones."""
        original_features = self._columns_to_encode
        encoded_features = self.get_feature_names_()

        num_features = [len(x) for x in self.categories_]
        num_features.insert(0, 0)

        category_idxs = np.cumsum(num_features).tolist()
        category_idxs = [
            (category_idxs[i], category_idxs[i + 1])
            for i in range(len(category_idxs) - 1)
        ]

        link = {
            col: col
            for col in self._input_columns
            if col not in self._columns_to_encode
        }
        for original_feature_idx, (encoded_start, encoded_end) in enumerate(
            category_idxs
        ):
            for encoded_feature_idx in range(encoded_start, encoded_end):
                link[encoded_features[encoded_feature_idx]] = original_features[
                    original_feature_idx
                ]

        return link

    @staticmethod
    def _get_unique_name(name, set_):
        if name not in set_:
            return name

        idx = 2
        while True:
            new_name = f"{name}_{idx}"
            if new_name not in set_:
                return new_name

            idx += 1

    def _validate_no_unexpected_categories(self, x):
        for idx, col in enumerate(self._columns_to_encode):
            expected_cats = set(self.categories_[idx])
            actual_cats = set(x[col].unique())

            if not actual_cats.issubset(expected_cats):
                unexpected_cats = actual_cats.difference(expected_cats)
                err_str = (
                    f"Found unexpected categories {unexpected_cats} in column {col}."
                )
                logger.error(err_str)
                raise OheUnexpectedLevelsException(err_str, (unexpected_cats, col))



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Definition of the DropNaRowsTransformer."""

import pandas as pd

import sklearn.base


class DropNaRowsTransformer(sklearn.base.TransformerMixin):
    """Custom transformer for dropping rows when a given column has a null value."""

    def __init__(self, colname, inplace=False):
        self.colname = colname
        self.inplace = bool(inplace)

    def fit(self, x, y=None, **kwargs):
        """Fit the transformer."""

        return self

    def transform(self, df):
        """Transform to remove NA rows."""
        inplace = self.inplace

        assert isinstance(df, pd.DataFrame)
        assert self.colname in list(df)

        result = df.dropna(subset=[self.colname], inplace=inplace)
        if not inplace:
            df = result

        return df



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.

"""Setup.py for ayx-learn."""
import os
from distutils.core import setup

from setuptools import find_packages

this_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

module_name = "ayx_learn"


def get_new_version(package):
    """Generate the next version of the package."""
    try:
        exec(open(os.path.join(package, "version.py")).read(), globals())

        # Get the package version from the current check in
        package_base_version_str = __version__  # noqa: F821
        package_split_str = package_base_version_str.split(".")

        package_major = int(package_split_str[0])
        package_minor = int(package_split_str[1])
        package_patch = int(package_split_str[2])

        pipeline_id = os.getenv("CI_PIPELINE_ID") or "0"

        new_version = ".".join(
            [
                str(package_major),
                str(package_minor),
                str(package_patch),
                str(pipeline_id),
            ]
        )

    except Exception as e:
        print(str(e))
        new_version = "0.0.0"

    print("New version: " + new_version)
    return new_version


version = get_new_version(module_name)

with open(os.path.join(module_name, "version.py"), "a") as f:
    f.write('\n__version__ = "%s"' % (version,))

description = "Alpha release of ayx-learn."

requirements = ["pandas==0.24.2", "scikit-learn==0.21.1"]

setup(
    name=module_name,
    version=version,
    author="Alteryx, Inc.",
    author_email="support@alteryx.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.alteryx.com/hamilton/ayx-learn",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=requirements,
)



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Definition of the ColumnTypeTransformer for applying data typing to a dataframe."""

from ayx_learn.utils.constants import ColumnTypes
from ayx_learn.utils.typing import typing_conversions
from ayx_learn.utils.validate import validate_col_type

import pandas as pd

import sklearn.base


class ColumnTypeTransformer(sklearn.base.TransformerMixin):
    """Transformer used for applying typing to a dataframe."""

    def __init__(
        self, coltype=ColumnTypes.NUMERIC, colname=None, inplace=False, **kwargs
    ):
        self.coltype = coltype
        self.colname = colname
        self._inplace = bool(inplace)

    @property
    def coltype(self):
        """Getter for column type of this transformer."""
        return self.__coltype

    @coltype.setter
    def coltype(self, value):
        """Setter for column type of this transformer."""
        validate_col_type(value)
        self.__coltype = value

    def transform(self, x, y=None, **kwargs):
        """Apply typing transforms to input dataframe."""
        if not isinstance(x, pd.DataFrame):
            raise TypeError("Input must be a pandas dataframe.")
        columns_to_transform = x.columns
        if self.colname is not None:
            columns_to_transform = [self.colname]
        inplace = self._inplace
        if not inplace:
            x = x.copy()
        # Apply transformation columnwise
        for column in columns_to_transform:
            x.loc[:, column] = typing_conversions[self.coltype](x[column])
        return x

    def fit(self, x, y=None, **kwargs):
        """Fit the transformer."""
        return self



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Validation utility functions."""
import logging
from enum import Enum
from typing import List

from ayx_learn.utils.constants import ColumnTypes
from ayx_learn.utils.exceptions import NullValueError

logger = logging.getLogger(__name__)


def validate_enum(value: Enum, enum: Enum):
    """Validate value is in the enum.

    Parameters
    ----------
    value : Enum
        Value to be tested
    enum : Enum
        Enum to be tested

    Returns
    -------
    None
        If enum is valid.

    Raises
    ------
    TypeError
        If value is not in enum.
    """
    if not isinstance(value, enum):
        raise TypeError(f"{value} is not a valid {enum}")


def validate_list_of_str(lst: List[str]):
    """Validate that list consists of elements of type str.

    Parameters
    ----------
    lst : List[str]
        List of type str.

    Returns
    -------
    None
        If valid list.

    Raises
    ------
    TypeError
        If not a List or not all elements of list are of type str.
    """
    if not isinstance(lst, List):
        raise TypeError(f"{lst} is not a valid {List}")

    if not all(isinstance(elem, str) for elem in lst):
        raise TypeError(f"{lst} is not all of type str")


def validate_col_type(coltype):
    """Assert coltype is valid."""
    return validate_enum(coltype, ColumnTypes)


def validate_no_nulls(df):
    """Assert that a dataframe contains no null values."""
    for col in list(df):
        if df[col].isnull().values.any():
            err_str = f"Dataframe contains null values in column: {col}"
            logger.error(err_str)
            raise NullValueError(err_str, col)



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Feature union for pandas dataframes in sklearn pipelines."""

import numpy as np

import pandas as pd

import scipy

import sklearn.pipeline


class DataFrameFeatureUnion(sklearn.pipeline.FeatureUnion):
    """
    Override of sklearn FeatureUnion for use with dataframes.

    The purpose of this class is to allow for using pandas DataFrames in sklearn
    Transformers and pipelines while preserving DataFrame features
    like column names, etc.

    See blog post here where this was copied from:
    https://zablo.net/blog/post/pandas-dataframe-in-scikit-learn-feature-union

    Arguments:
        transformer_list: list -- List of (string, transformer) tuples

        n_jobs: int  -- Number of jobs to run in parallel

        transformer_weights : dict, optional -- Multiplicative weights
            for features per transformer.  Keys are transformer names,
            values the weights.
    """

    def _find_duplicates(self, lst):
        found = set()
        duplicates = set()
        for value in lst:
            if value in found:
                duplicates.add(value)
            found.add(value)

        return list(duplicates)

    def _merge_dataframes_by_column(self, xs):
        retval = pd.concat(xs, axis="columns", copy=False)
        col_names = list(retval)

        duplicate_cols = self._find_duplicates(col_names)
        if len(duplicate_cols) > 0:
            raise ValueError(f"Duplicate columns found: f{duplicate_cols}")

        return retval

    def transform(self, x, y=None):
        """Apply transform to each parallel step in the pipeline and feature union."""
        xs = []
        for _, trans, weight in self._iter():
            xs.append(trans.transform(x))

        if not xs:
            # All transformers are None
            return np.zeros((x.shape[0], 0))
        if any(scipy.sparse.issparse(f) for f in xs):
            xs = scipy.sparse.hstack(xs).tocsr()
        else:
            xs = self._merge_dataframes_by_column(xs)
        return xs

    def fit(self, x, y=None):
        """Fit each transform in the feature union."""
        self.transformer_list = list(self.transformer_list)
        self._validate_transformers()
        transformers = []
        for _, trans, _ in self._iter():
            transformers.append(trans.fit(x, y))
        self._update_transformer_list(transformers)
        return self

    def fit_transform(self, x, y=None, **fit_params):
        """
        Fit/transform the steps in the feature union.

        Override of sklearn.FeatureUnion fit_transform
        that handles pandas DataFrame objects, ensuring
        that output will still be a pandas DataFrame.

        Arguments:
            x: pandas DataFrame object -- Pandas DataFrame
            containing the data columns specified in the
            a PandasTransform constructor.

        Keyword Arguments:
            y: pandas Series object -- Targets for supervised learning (default: {None})

        Returns
        -------
            pandas.DataFrame -- A pandas DataFrame containing the
            results of the transformations specified.
        """
        self.fit(x, y)
        return self.transform(x, y)



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Column select and column drop data frame transformer."""

import re

from ayx_learn.utils.validate import validate_list_of_str

import sklearn.base


class ColumnSelectorTransformer(sklearn.base.TransformerMixin):
    """Transformer used to select/deselect a set of columns based on regex.

    Column Selector/Deselector: Takes a list of regular expressions
    to match column names, and a flag to indicate if you want to select
    or deselect those columns.

    """

    def __init__(
        self, col_select=None, select_not_deselect=True, inplace=False, **kwargs
    ):
        if col_select is not None:
            validate_list_of_str(col_select)

        if col_select is None:
            self._col_select_regexs = []
        else:
            self._col_select_regexs = [re.compile(str(exp)) for exp in list(col_select)]

        self._select_not_deselect = select_not_deselect
        self._inplace = bool(inplace)

    def fit(self, x, y=None, **kwargs):
        """Fit the transformer. Noop in this case."""
        return self

    def transform(self, x, y=None, copy=False, **kwargs):
        """Select columns from dataframe based on regex."""
        inplace = self._inplace
        # Identify columns to select/deselect
        col_select = [list(filter(r.match, list(x))) for r in self._col_select_regexs]
        # Flatten the array
        col_select = [item for row in col_select for item in row]
        if self._select_not_deselect:
            result = x.drop(x.columns.difference(col_select), 1, inplace=inplace)
        else:
            result = x.drop(col_select, 1, inplace=inplace)
        if not inplace:
            x = result
        return x
